import contextlib
import copy
import functools
import time
from collections import Counter
from collections import namedtuple

_cache = {}

config = {
    'cache_time': 10,
    'ignore_cache': False,
}

default_stats = Counter({
    'expirations': 0,
    'sets': 0,
    'gets': 0,
    'hits': 0,
    'misses': 0
})

stats = Counter()

Key = namedtuple('Key', ['original_key'])
Record = namedtuple('Record', ['expiry', 'value'])


def configure(cache_time=config['cache_time']):
    """Configure the cache and reset its values"""
    config['cache_time'] = cache_time
    stats.clear()
    stats.update(default_stats)
    _cache.clear()


def has_expired(record, now):
    if record.expiry < now:
        return True
    else:
        return False


def getv(key, now=None):
    """Get a key from the cache

    :param now: The current time
    :raises: KeyError if the key is not present or has expired
    :returns: The result
    """
    if now is None:
        now = time.time()
    key = Key(key)
    stats['gets'] += 1
    if key in _cache:
        record = _cache[key]
        if has_expired(record, now) or config['ignore_cache']:
            stats['expirations'] += 1
            del _cache[key]
        else:
            stats['hits'] += 1
            return record.value
    stats['misses'] += 1
    raise KeyError(key)


def setv(key, value):
    key = Key(key)
    stats['sets'] += 1
    expiration_time = time.time() + config['cache_time']
    rec = Record(expiration_time, value)
    _cache[key] = rec


def get_stats():
    return copy.copy(stats)


@contextlib.contextmanager
def maybe_bust(bust_or_not):
    previous_state = config['ignore_cache']
    config['ignore_cache'] = bust_or_not
    yield
    config['ignore_cache'] = previous_state


def cached(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        now = time.time()
        key = tuple([func.__name__, args])
        try:
            response = getv(key, now)
        except KeyError:
            response = func(*args, **kwargs)
            setv(key, response)
        return response
    return wrapper



#!/usr/bin/env python

from __future__ import print_function

import optparse
import sys

import hacheck.spool


def up():
    return main('up')


def down():
    return main('down')


def status():
    return main('status')


def status_all():
    return main('status_all')


def print_s(fmt_string, *formats):
    """Print function split out for mocking"""
    print(fmt_string % formats)


def main(default_action='status'):
    ACTIONS = ('up', 'down', 'status', 'status_all')
    parser = optparse.OptionParser(usage='%prog [options] service_name')
    parser.add_option('--spool-root', default='/var/spool/hacheck',
        help='Root for spool for service states (default %default)')
    parser.add_option('-a', '--action', type='choice', choices=ACTIONS, default=default_action,
        help='Action (one of %s, default %%default)' % ', '.join(ACTIONS, ))
    parser.add_option('-r', '--reason', type=str, default="", help='Reason string when setting down')
    opts, args = parser.parse_args()

    if opts.action != 'status_all':
        if len(args) != 1:
            parser.error('Wrong number of arguments')
        service_name = args[0]

    hacheck.spool.configure(opts.spool_root, needs_write=True)

    if opts.action == 'up':
        hacheck.spool.up(service_name)
        return 0
    elif opts.action == 'down':
        hacheck.spool.down(service_name, opts.reason)
        return 0
    elif opts.action == 'status_all':
        for service_name, info in hacheck.spool.status_all_down():
            print_s('DOWN\t%s\t%s', service_name, info.get('reason', ''))
        return 0
    else:
        status, info = hacheck.spool.status(service_name)
        if status:
            print_s('UP\t%s', service_name)
            return 0
        else:
            print_s('DOWN\t%s\t%s', service_name, info.get('reason', ''))
            return 1


if __name__ == '__main__':
    sys.exit(main())



import datetime
import socket
import time

import tornado.concurrent
import tornado.ioloop
import tornado.gen
import tornado.httpclient

from . import cache
from . import spool
from . import __version__

TIMEOUT = 10


# Do not cache spool checks
@tornado.concurrent.return_future
def check_spool(service_name, port, query, io_loop, callback):
    up, extra_info = spool.is_up(service_name)
    if not up:
        info_string = 'Service %s in down state' % (extra_info['service'],)
        if extra_info.get('reason', ''):
            info_string += ": %s" % extra_info['reason']
        callback((503, info_string))
    else:
        callback((200, extra_info.get('reason', '')))


# IMPORTANT: the gen.coroutine decorator needs to be the innermost
@cache.cached
@tornado.gen.coroutine
def check_http(service_name, port, query, io_loop):
    if not query.startswith("/"):
        query = "/" + query  # pragma: no cover
    request = tornado.httpclient.HTTPRequest('http://127.0.0.1:%d%s' % (port, query), method='GET',
            headers={'User-Agent': 'hastate %s' % (__version__)}, request_timeout=TIMEOUT)
    http_client = tornado.httpclient.AsyncHTTPClient(io_loop=io_loop)
    try:
        response = yield http_client.fetch(request)
        code = response.code
        reason = response.body
    except tornado.httpclient.HTTPError as exc:
        code = exc.code
        reason = exc.response.body if exc.response else ""
    raise tornado.gen.Return((code, reason))


@cache.cached
@tornado.gen.coroutine
def check_tcp(service_name, port, query, io_loop):
    stream = None
    connect_start = time.time()

    def timed_out():
        try:
            stream.close()
        except:  # pragma: no cover
            pass
        raise tornado.gen.Return((503, 'Connection timed out after %.2fs' % (time.time() - connect_start)))

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
    stream = tornado.iostream.IOStream(s, io_loop=io_loop)
    timeout = io_loop.add_timeout(datetime.timedelta(seconds=TIMEOUT), timed_out)
    yield tornado.gen.Task(stream.connect, ("127.0.0.1", port))
    io_loop.remove_timeout(timeout)
    stream.close()
    raise tornado.gen.Return((200, 'Connected in %.2fs' % (time.time() - connect_start)))



import os

config = {
    'spool_root': None,
}


def configure(spool_root, needs_write=False):
    access_required = os.W_OK | os.R_OK if needs_write else os.R_OK
    if os.path.exists(spool_root):
        if not os.access(spool_root, access_required):
            raise ValueError("Insufficient access to %s" % spool_root)
    else:
        os.mkdir(spool_root, 0o750)
    config['spool_root'] = spool_root


def is_up(service_name):
    """Check whether a service is asserted to be up or down. Includes the logic
    for checking system-wide all state

    :returns: (bool of service status, dict of extra information)
    """
    all_up, all_info = status("all")
    if all_up:
        return status(service_name)
    else:
        return all_up, all_info


def status(service_name):
    """Check whether a service is asserted to be up or down, without checking
    the system-wide 'all' state.

    :returns: (bool of service status, dict of extra information)
    """
    try:
        with open(os.path.join(config['spool_root'], service_name), 'r') as f:
            reason = f.read()
            return False, {'service': service_name, 'reason': reason}
    except IOError:
        return True, {'service': service_name, 'reason': ''}


def status_all_down():
    """List all down services

    :returns: Iterable of pairs of (service name, dict of extra information)
    """
    for service_name in os.listdir(config['spool_root']):
        up, info = status(service_name)
        if not up:
            yield service_name, info


def up(service_name):
    try:
        os.unlink(os.path.join(config['spool_root'], service_name))
    except OSError:
        pass


def down(service_name, reason=""):
    with open(os.path.join(config['spool_root'], service_name), 'w') as f:
        f.write(reason)



import logging
import time

import tornado.ioloop
import tornado.httputil
import tornado.httpclient
import tornado.gen
import tornado.web

from . import cache
from . import checker

log = logging.getLogger('hacheck')


class StatusHandler(tornado.web.RequestHandler):
    def get(self):
        stats = {}
        stats['cache'] = cache.get_stats()
        stats['uptime'] = time.time() - self.settings['start_time']
        self.set_status(200)
        self.write(stats)


class BaseServiceHandler(tornado.web.RequestHandler):
    CHECKERS = []

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self, service_name, port, query):
        with cache.maybe_bust(self.request.headers.get('Pragma', '') == 'no-cache'):
            port = int(port)
            last_message = ""
            for checker in self.CHECKERS:
                code, message = yield checker(service_name, port, query, io_loop=tornado.ioloop.IOLoop.current())
                last_message = message
                if code > 200:
                    if code in tornado.httputil.responses:
                        self.set_status(code)
                    else:
                        self.set_status(503)
                    self.write(message)
                    self.finish()
                    break
            else:
                self.set_status(200)
                self.write(last_message)
                self.finish()


class SpoolServiceHandler(BaseServiceHandler):
    CHECKERS = [checker.check_spool]


class HTTPServiceHandler(BaseServiceHandler):
    CHECKERS = [checker.check_spool, checker.check_http]


class TCPServiceHandler(BaseServiceHandler):
    CHECKERS = [checker.check_spool, checker.check_tcp]



import logging
import optparse
import signal
import time
import sys

import tornado.ioloop
import tornado.web

from . import cache
from . import handlers
from . import spool


def get_app():
    return tornado.web.Application([
        (r'/http/([a-zA-Z0-9_-]+)/([0-9]+)/(.*)', handlers.HTTPServiceHandler),
        (r'/tcp/([a-zA-Z0-9_-]+)/([0-9]+)/?(.*)', handlers.TCPServiceHandler),
        (r'/spool/([a-zA-Z0-9_-]+)/([0-9]+)/?(.*)', handlers.SpoolServiceHandler),
        (r'/status', handlers.StatusHandler),
    ], start_time=time.time())


def main():
    parser = optparse.OptionParser()
    parser.add_option('-p', '--port', default=3333, type=int)
    parser.add_option('--cache-time', default=10.0, type=float,
            help='How many seconds to cache response for (default %default)')
    parser.add_option('--spool-root', default='/var/spool/hacheck',
            help='Root for spool for service states (default %default)')
    parser.add_option('-v', '--verbose', default=False, action='store_true')
    opts, args = parser.parse_args()
    logging.basicConfig(level=(logging.DEBUG if opts.verbose else logging.WARNING))
    cache.configure(cache_time=opts.cache_time)
    spool.configure(spool_root=opts.spool_root)
    application = get_app()
    application.listen(opts.port)
    ioloop = tornado.ioloop.IOLoop.instance()
    for sig in (signal.SIGTERM, signal.SIGQUIT, signal.SIGINT):
        signal.signal(sig, lambda *args: ioloop.stop())
    ioloop.start()
    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
from setuptools import setup, find_packages

import hacheck


with open('requirements.txt') as f:
    required = f.read().splitlines()


setup(
    name="hacheck",
    version=hacheck.__version__,
    author="James Brown",
    author_email="jbrown@uber.com",
    url="https://github.com/uber/hacheck",
    license="MIT",
    packages=find_packages(exclude=['tests']),
    keywords=["monitoring", "load-balancing", "networking"],
    description="HAProxy health-check proxying service",
    install_requires=required,
    test_suite="nose.collector",
    entry_points={
        'console_scripts': [
            'haup = hacheck.haupdown:up',
            'hadown = hacheck.haupdown:down',
            'hastatus = hacheck.haupdown:status',
            'hacheck = hacheck.main:main',
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Topic :: System :: Monitoring",
    ]
)
